#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef _OPENMP
#include <omp.h>
#endif

double getClock();

double **new_matrix(size_t rows, size_t cols);
void delete_matrix(double **mat, size_t rows);
double **rand_matrix(double **mat, size_t rows, size_t cols);
double checksum_matrix(double **mat, size_t rows, size_t cols);

/*   Matrix stored as an array of pointers to arrays containing the rows
                m
                |
                v
              { a,          b,          c }
a -> 1 0 0      |           |           |
b -> 0 2 0      v           |           |
b -> 0 0 3 => { 1, 0, 0 }   v           |
                          { 0, 2, 0 }   v
                                      { 0, 0, 3 } => m[1][1] == b[1] == 2
                                                   ^         == (*m[1])[1]
    The row pointer array enables this syntax -----+         == *(*(m+1)+1)
    Note that neither a[1*3+1] nor *(m[0]+3+1)                   ^
    are valid since MEMORY IS NOT CONTIGUOUS! -------------------+
*/

// C (m x n) = A (m x p) * B (p x n)
void matmul(size_t m, size_t n, size_t p, double **A, double **B, double **C) { 
    // Initialization
    for (size_t i = 0; i < m; i++) {
        for (size_t j = 0; j < n; j++) {
            C[i][j] = 0;
        }
    }

    // Accumulation
    for (size_t i = 0; i < m; i++) {
        for (size_t j = 0; j < n; j++) {
            for (size_t k = 0; k < p; k++) {
                C[i][j] += A[i][k] * B[k][j];
            }
        }
    }
}

int main(int argc, char *argv[]) {
    int param_iters = 1;

    if (argc != 2) {
        printf("Usage: %s <n> \n", argv[0]);
        printf("  <n> is the desired test size.\n");
        return 1;
    }

    // Reads the test parameters from the command line
    int param_n = 0;
    sscanf(argv[1], "%d", &param_n);
    printf("- Input parameters\n");
    printf("n\t= %i\n", param_n);
    size_t rows = param_n, cols = param_n;

    // Allocates input/output resources
    double **in1_mat = new_matrix(rows, cols);
    double **in2_mat = new_matrix(rows, cols);
    double **out_mat = new_matrix(rows, cols);
    if (!in1_mat || !in2_mat || !out_mat) {
        printf("Error: not enough memory to run the test using n = %i\n", param_n);
        return 1;
    }

    // Initializes data
    rand_matrix(in1_mat, rows, cols);
    rand_matrix(in2_mat, rows, cols);

    // Calls to the corresponding function to perform the computation
    printf("- Executing test...\n");
    double time_start = getClock();
    // ================================================

    for (int iters = 0; iters < param_iters; iters++) {
        matmul(
            rows, cols, cols,
            in1_mat, in2_mat, out_mat);
    }

    // ================================================
    double time_finish = getClock();

    // Prints an execution report
    double checksum = checksum_matrix(out_mat, rows, cols);
    printf("time (s)= %.6f\n", time_finish - time_start);
    printf("size\t= %i\n", param_n);
    printf("chksum\t= %.0f\n", checksum);
    if (param_iters > 1)
        printf("iters\t= %i\n", param_iters);

    // Release allocated resources
    delete_matrix(in1_mat, rows);
    delete_matrix(in2_mat, rows);
    delete_matrix(out_mat, rows);

    return 0;
}

double getClock() {
#ifdef _OPENMP
    return omp_get_wtime();
#elif __linux__ || __APPLE__
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return ts.tv_sec + ts.tv_nsec / 1.0e9;
#else
    // Warning: this clock is invalid for parallel applications
    return (double)clock() / CLOCKS_PER_SEC;
#endif
}

// Creates a new dense matrix with the specified rows and columns
double **new_matrix(size_t rows, size_t cols) {
    if (rows < 1 || cols < 1)
        return NULL;

    // Allocate an array of pointers to store the beginning of each row
    double **mat = (double **)calloc(rows, sizeof(double*));
    if (!mat)
        return NULL;

    // Allocate an array of doubles per row 
    for (size_t i = 0; i < rows; i++) {
        mat[i] = (double *)malloc(cols * sizeof(double));

        if (!mat[i]) { // Failed: free allocated memory and return
            for (size_t j = 0; j < i; j++)
                free(mat[j]);
            free(mat);
            return NULL;
        }
    }

    return mat;
}

// Deletes the matrix and the resources allocated by it
void delete_matrix(double **mat, size_t rows) {
    if (mat) {
        for (size_t i = 0; i < rows; i ++)
            free(mat[i]);
        free(mat);
    }
}

// Generates a random dense matrix
double **rand_matrix(double **mat, size_t rows, size_t cols) {
    if (!mat)
        return NULL;

    for (size_t row = 0; row < rows; row++)
        for (size_t col = 0; col < cols; col++)
            mat[row][col] = rand() % 10;
    return mat;
}

// Generates a checksum based on the matrix data
double checksum_matrix(double **mat, size_t rows, size_t cols) {
    if (!mat)
        return 0.0;

    double checkSum = 0.0;
    for (size_t row = 0; row < rows; row++)
        for (size_t col = 0; col < cols; col++)
            checkSum += mat[row][col];
    return checkSum;
}
