#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef _OPENMP
#include <omp.h>
#endif

double getClock();

/*  Matrix stored as a linearized array dynamically allocated
              mat
    1 0 0      |
    0 2 0      v
    0 0 3 => { 1, 0, 0, 0, 2, 0, 0, 0, 3 } => mat[1 * 3 + 1] == 2
*/

// C (m x n) = A (m x p) * B (p x n)
void matmul(size_t m, size_t n, size_t p, double *A, double *B, double *C) { 
    // Initialization
    for (size_t i = 0; i < m; i++) {
        for (size_t j = 0; j < n; j++) {
            C[i * n + j] = 0;
        }
    }
    // Since accessing by row and columm is irrelevant we could have written:
    // for (size_t i = 0; i < m * n; i++)
    //     C[i] = 0;

    // Accumulation
    for (size_t i = 0; i < m; i++) {
        for (size_t j = 0; j < n; j++) {
            for (size_t k = 0; k < p; k++) {
                C[i * n + j] += A[i * p + k] * B[k * n + j];
                // Equivalent to C[i][j] += A[i][k] * B[k][j]
            }
        }
    }
}

int main(int argc, char *argv[]) {
    int param_iters = 1;

    if (argc != 2) {
        printf("Usage: %s <n> \n", argv[0]);
        printf("  <n> is the desired test size.\n");
        return 1;
    }

    // Reads the test parameters from the command line
    int param_n = 0;
    sscanf(argv[1], "%d", &param_n);
    printf("- Input parameters\n");
    printf("n\t= %i\n", param_n);
    size_t rows = param_n, cols = param_n;

    // Allocates input and output matrices as dynamic one-dimensional arrays
    double *in1_mat = (double *)malloc(rows * cols * sizeof(double));
    double *in2_mat = (double *)malloc(rows * cols * sizeof(double));
    double *out_mat = (double *)malloc(rows * cols * sizeof(double));
    if (!in1_mat || !in2_mat || !out_mat) {
        printf("Error: not enough memory to run the test using n = %i\n", param_n);
        return 1;
    }

    // Initializes data (two loops because the rand calling order matters)
    for (size_t i = 0; i < rows * cols; i++)
        in1_mat[i] = rand() % 10;
    for (size_t i = 0; i < rows * cols; i++)
        in2_mat[i] = rand() % 10;

    // Calls to the corresponding function to perform the computation
    printf("- Executing test...\n");
    double time_start = getClock();
    // ================================================

    for (int iters = 0; iters < param_iters; iters++) {
        matmul(
            rows, cols, cols,
            in1_mat, in2_mat, out_mat);
    }

    // ================================================
    double time_finish = getClock();

    // Compute the checksum
    double checksum = 0.0;
    for (size_t i = 0; i < rows * cols; i++)
        checksum += out_mat[i];

    // Prints an execution report
    printf("time (s)= %.6f\n", time_finish - time_start);
    printf("size\t= %i\n", param_n);
    printf("chksum\t= %.0f\n", checksum);
    if (param_iters > 1)
        printf("iters\t= %i\n", param_iters);

    // Release allocated resources
    free(in1_mat);
    free(in2_mat);
    free(out_mat);

    return 0;
}

double getClock() {
#ifdef _OPENMP
    return omp_get_wtime();
#elif __linux__ || __APPLE__
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return ts.tv_sec + ts.tv_nsec / 1.0e9;
#else
    // Warning: this clock is invalid for parallel applications
    return (double)clock() / CLOCKS_PER_SEC;
#endif
}
